import tempfile
import os
import subprocess
import re

from switss.prism import prism_to_tra, parse_label_file

from config import comics_path, timeout

class ParseError(Exception):
    pass

def write_header(csv_file_path):
    try:
        os.mkdir(os.path.dirname(csv_file_path))
    except:
        pass
    with open(csv_file_path,"w") as csv_file:
        csv_file.write("tool,mode,threshold,t_wall,t_cpu,timeout,mc time,states,lower_bound,trans,prob,prob LP,heur_iter,mem_max,info\n")

def l(something):
    if something == None:
        return "-"
    else:
        return str(something)

def write_line(csv_file_path,tool=None,mode=None,thr=None,t_wall=None,t_cpu=None,timeout=None,mc_time=None,states=None,lower_bound=None,trans=None,prob=None,prob_lp=None,info=None,heur_iter=None,mem_max=None):
    with open(csv_file_path,"a") as csv_file:
        csv_file.write(l(tool) + "," + l(mode) + "," + l(thr) + "," + l(t_wall) + "," + l(t_cpu) + "," + l(timeout) + "," + l(mc_time) + "," + l(states) + "," + l(lower_bound) + "," + l(trans) + "," + l(prob) + "," + l(prob_lp) + "," + l(heur_iter) + "," + l(mem_max) + "," + l(info) + "\n")

def prism_to_comics(model_file_path,target_label,dest_path,prism_constants = {}, extra_labels = {}):
    with tempfile.TemporaryDirectory() as tempdirname:
        temp_model_file = os.path.join(tempdirname, "model")
        tra_path = temp_model_file + ".tra"
        lab_path = temp_model_file + ".lab"
        if not prism_to_tra(model_file_path,temp_model_file,prism_constants=prism_constants, extra_labels=extra_labels):
            assert False, "Prism call to create model failed"

        states_by_label,_,_ = parse_label_file(lab_path)
        assert len(states_by_label["init"]) == 1
        init = list(states_by_label["init"])[0]
        target_states = states_by_label[target_label]

        with open(tra_path) as tra_file:
            with open(dest_path,"w") as dest_file:
                for line in tra_file:
                    line_split = line.split()

                    # check for first lines, which has format "#states #transitions"
                    if len(line_split) == 2:
                        N = int(line_split[0])
                        T = int(line_split[1])
                        dest_file.write("STATES " + str(N) + "\n")
                        dest_file.write("TRANSITIONS " + str(N) + "\n")
                        dest_file.write("INITIAL " + str(init) + "\n")
                        for s in target_states:
                            dest_file.write("TARGET " + str(s) + "\n")
                        # all other lines have format "from to prob"
                    else:
                        dest_file.write(line)

def write_comics_conf(conf_path,dtmc_path,mode,threshold):
    with open(conf_path,"w") as conf_file:
        conf_file.write("TASK Counterexample\n")
        conf_file.write("PROBABILITY_BOUND " + str(threshold) + "\n")
        conf_file.write("DTMC_FILE " + os.path.relpath(dtmc_path,os.path.dirname(conf_path)) + "\n")
        conf_file.write("SEARCH_ALGORITHM " + mode + "\n")
        conf_file.write("ABSTRACTION concrete\n")

def parse_comics_result(res_path):
    found_size = False
    found_origin_prob = False
    found_subsys_prob = False
    found_mc_time = False
    found_ce_time = False
    subsys_regexp = re.compile("Counter example size: ([0-9]+) states, ([0-9]+) transitions")
    orig_prob_regexp = re.compile("Model Checking result of original system: ([+-]?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?)")
    subsys_prob_regexp = re.compile("Closure probability: ([+-]?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?)")
    mc_time_regexp = re.compile("Time for model checking: ([+-]?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?) secs")
    ce_time_regexp = re.compile("Time of counter example computation: ([+-]?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?) secs")
    with open(res_path) as res_file:
        for line in res_file:
            match = re.match(subsys_regexp,line)
            if match:
                subsys_states = int(match.group(1))
                subsys_trans = int(match.group(2))
                found_size = True
                continue
            match = re.match(orig_prob_regexp,line)
            if match:
                orig_prob = float(match.group(1))
                found_origin_prob = True
                continue
            match = re.match(subsys_prob_regexp,line)
            if match:
                subsys_prob = float(match.group(1))
                found_subsys_prob = True
                continue
            match = re.match(mc_time_regexp,line)
            if match:
                mc_time = float(match.group(1))
                found_mc_time = True
                continue
            match = re.match(ce_time_regexp,line)
            if match:
                ce_time = float(match.group(1))
                found_ce_time = True
                break
    if found_size and found_origin_prob and found_subsys_prob and found_mc_time and found_ce_time:
        return subsys_states,subsys_trans,orig_prob,subsys_prob,mc_time,ce_time
    else:
        raise ParseError("couldnt parse comics result")


def run_comics(model_path,csv_file_path,thr):
    if not os.path.isfile(model_path):
        return

    # run comics with global search strategy
    comics_conf_file = tempfile.NamedTemporaryFile()
    rel_path_conf = os.path.relpath(comics_conf_file.name)

    write_comics_conf(rel_path_conf,os.path.abspath(model_path),"global",thr)

    print(rel_path_conf)
    print(model_path)
    print(comics_conf_file.name)
    print(comics_conf_file.read())

    try:
        print((subprocess.check_output([comics_path,rel_path_conf],timeout=timeout)).decode("utf-8")) #
    except subprocess.TimeoutExpired:
        write_line(csv_file_path,tool="comics",mode="global",thr=thr,timeout=True)
    except subprocess.CalledProcessError as e:
        print("Warning: comics returned with returncode" + str(e.returncode))
        print(e.output)
        with open("error.log","a") as error_file:
            error_file.write(e.output.decode("utf-8"))
        if e.returncode == 6:
            write_line(csv_file_path,tool="comics",mode="global",thr=thr,info="MEM-OUT")
        else:
            write_line(csv_file_path,tool="comics",mode="global",thr=thr,info="error")
    else:
        try:
            states,trans,orig_prob,sub_prob,mc_time,ce_time = parse_comics_result("counter_example_summary.txt")
        except ParseError:
            write_line(csv_file_path,tool="comics",mode="global",thr=thr,info="parse error")
            print("Warning: comics counter_example_summary.txt could not be parsed!")
            with open("counter_example_summary.txt") as ce_file:
                with open("error.log", "a") as error_file:
                    print("Warning: comics counter_example_summary.txt could not be parsed!")
                    error_file.write("Warning: comics counter_example_summary.txt could not be parsed!")
                    for line in ce_file:
                        print(line)
                        error_file.write(line)
        else:
            write_line(csv_file_path,tool="comics",mode="global",thr=thr,t_wall=ce_time,states=states,prob=sub_prob)
            os.remove("counter_example_summary.txt")

    # run comics with local search strategy
    write_comics_conf(rel_path_conf,os.path.abspath(model_path),"local",thr)
    try:
        print(rel_path_conf)
        with open(rel_path_conf) as conf_file:
            for line in conf_file:
                print(line)
        subprocess.check_output([comics_path,rel_path_conf],timeout=timeout)
    except subprocess.TimeoutExpired:
        write_line(csv_file_path,tool="comics",mode="local",thr=thr,timeout=True)
    except subprocess.CalledProcessError as e:
        print("Warning: comics returned with returncode" + str(e.returncode))
        print(e.output)
        with open("error.log","a") as error_file:
            error_file.write(e.output.decode("utf-8"))
        if e.returncode == 6:
            write_line(csv_file_path,tool="comics",mode="local",thr=thr,info="MEM-OUT")
        else:
            write_line(csv_file_path,tool="comics",mode="local",thr=thr,info="error")
    else:
        try:
            states,trans,orig_prob,sub_prob,mc_time,ce_time = parse_comics_result("counter_example_summary.txt")
        except ParseError:
            write_line(csv_file_path,tool="comics",mode="local",thr=thr,info="parse error")
            print("Warning: comics counter_example_summary.txt could not be parsed!")
            with open("counter_example_summary.txt") as ce_file:
                with open("error.log","a") as error_file:
                    print("Warning: comics counter_example_summary.txt could not be parsed!")
                    error_file.write("Warning: comics counter_example_summary.txt could not be parsed!")
                    for line in ce_file:
                        print(line)
                        error_file.write(line)
        else:
            write_line(csv_file_path,tool="comics",mode="local",thr=thr,t_wall=ce_time,states=states,prob=sub_prob)
            os.remove("counter_example_summary.txt")

    comics_conf_file.close()
